EPROM_START    equ $F000
RAM_START      equ $4000
STACK_START    equ $7FFF
RESET_VEC      equ $FFFE
FAIL_ADDR      equ $FF00

; Start of EPROM
            org  EPROM_START
start:      movb #8,$16
            movb #4,$0a
            ldx  #RAM_START
            lds  #STACK_START
            bra  run_prog
            
run_prog:   leas -7,SP
            movw #$4049,0,SP
            movw #$0fdb,2,SP
            movw #$4049,4,SP
            movw #$0fdb,6,SP
            jsr  fdiv
            
while_1:    bra while_1       ; loop forever

; fabs_sign returns with bit 0 = 1 if the first float (X+4) is negative and
; bit 1 = 1 if the second float (X) is negative. Both floats are replaced
; with their absolute values.
fabs_sign:
            clra
            ldab 4,X
            tstb
            bge  fabs_flt1
            adda #1
fabs_flt1:  andb #$7f
            stab 4,X
            ldab 0,X
            tstb
            bge  fabs_flt2
            adda #2
fabs_flt2:  andb #$7f
            stab 0,X
            rts

; fcomp returns "01" if the first float (X+4) is greater, "10" if the
; second float (X) is greater, or "00" if they are equal.
fcomp:
            clra
            ldd  4,X
            cpd  0,X
            bgt  flt1_more
            blt  flt2_more
            ldd  6,X
            cpd  2,X
            bgt  flt1_more
            blt  flt2_more
flt12_equ:  rts
flt1_more:  ldaa #1
            rts
flt2_more:  ldaa #2
            rts

; Convert a standard IEEE 754 float into an unsigned float, where the
; top byte is the exponent and the other 3 bytes are the fraction. Bit
; 23 is 0.
fproc:
            ldd  0,X          ; Load top 2 bytes of the float at X
            lsld              ; Shift D left 1 bit. A contains exponent.
            lsrb              ; Restore correct fractional bit positions.
            orab #$80         ; Set bit 23 of the float in B to 1.
            ;andb #$7f
            std  0,X          ; Save the top 2 bytes back to the original float.
            rts               ; Return to caller

; Convert a processed float back into an IEEE 754 float with no sign bit.            
funproc:
            ldd  0,X          ; Load top 2 bytes of the float at X
            lslb              ; Shift B left 1 bit to put fraction next to exponent.
            lsrd              ; Shift D right 1 bit to restore IEEE 754 bit positions.
            std  0,X          ; Save the top 2 bytes back to the original float.
            rts               ; Return to caller.

; Denormalizes the 2nd float (at X) to match the exponent of the 1st
; float (at X+4). The 2nd float is guaranteed to be smaller when this
; subroutine is run.
fdenorm:
            ldaa 4,X
            suba 0,X
            cmpa #24
            bge  fdn_zero
fdn_if0:    tsta
            beq  fdn_end
            cmpa #8
            blt  fdn_sr1
            cmpa #16
            blt  fdn_sr8
fdn_sr16:   movb 1,X,3,X
            clr  1,X
            clr  2,X
            suba #16
            bra  fdn_if0
fdn_sr8:    movw 1,X,2,X
            clr  1,X
            suba #8
            bra  fdn_if0
fdn_sr1:    lsr  1,X
            ror  2,X
            ror  3,X
            deca
            bra  fdn_if0
fdn_zero:   movw #0,0,X
            movb #0,2,X
fdn_end:    movb 4,X,0,X
            rts            

; Renormalizes the float at X to make sure the most significant bit
; falls on bit 23 of the float.
frenorm:    ldd  1,X
            cpd  #0
            blt  frn_end
            beq  frn_sl16
            tsta
            beq  frn_sl8
frn_sl1:    lsl  3,X
            rol  2,X
            rol  1,X
            dec  0,X
            bra  frenorm
frn_sl8:    movw 2,X,1,X
            clr  3,X
            ldaa 0,X
            suba #8
            staa 0,X
            bra  frenorm
frn_sl16:   ldaa 3,X
            tsta
            beq  frn_zero
            movb 3,X,1,X
            clr  2,X
            clr  3,X
            ldaa 0,X
            suba #16
            staa 0,X
            bra  frenorm
frn_zero:   movw #0,0,X
            movw #0,2,X
frn_end:    rts            

; Negate the float value at X. 
fnegate:
            ldaa 0,X          ; Load highest byte of float
            eora #$80         ; Flip sign bit by XORing with 1
            staa 0,X          ; Save highest byte of float to new value
            rts               ; Return to caller

; Set the float value at X to specifically be negative.
fsetneg:
            ldaa 0,X          ; Load highest byte of float
            oraa #$80         ; Set sign bit to 1
            staa 0,X          ; Set highest byte of float to new value
            rts               ; Return to caller


; Adds 2 processed floats. Processed floats have the following properties:
; Bits [31:24] are the exponent.
; Bit [23] = 0
; Bits [22:0] are the fraction.
add_procf:
            jsr  fdenorm
            ldd  6,X
            addd 2,X
            std  2,X
            ldaa 5,X
            adca 1,X
            staa 1,X
            bsr  frenorm
            rts

; Subtracts 2 processed floats (same properties as above)
sub_procf:
            jsr  fdenorm
            ldd  6,X
            subd 2,X
            std  2,X
            ldaa 5,X
            sbca 1,X
            staa 1,X
            jsr  frenorm
            rts
            
; Processes and adds 2 floats, which are pushed as arguments
; onto the stack.
fadd:
            leas -2,SP        ; Make 2 bytes space for local variables
            leax 4,SP         ; Set X to point to the first float
            jsr  fabs_sign    ; Get signs and set absolute values for floats
            staa 0,SP         ; Store sign information in [SP+0]
            jsr  fcomp        ; Determine which of these floats is greater
            staa 1,SP         ; Store comparison info in [SP+1]
            cmpa #2           ; if !(flt1 > flt2), order is correct
            bne  fadd_proc    ; skip swapping the floats
            
fadd_swap:  leas -4,SP        ; Make 4 bytes space for swapping
            movw 0,X,0,SP     ; Move top 2 bytes of flt2 to swap
            movw 2,X,2,SP     ; Move bottom 2 bytes of flt2 to swap
            movw 4,X,0,X      ; Move top 2 bytes of flt1 to flt2
            movw 6,X,2,X      ; Move bottom 2 bytes of flt1 to flt2
            movw 0,SP,4,X     ; Move top 2 bytes of swap to flt1
            movw 2,SP,6,X     ; Move bottom 2 bytes of swap to flt1
            leas 4,SP         ; Delete the temporary swap space
            
fadd_proc:  jsr  fproc        ; Process 2nd float (delete sign, align exponent)
            leax 4,X          ; Set X to point to 1st float
            jsr  fproc        ; Process 1st float (delete sign, align exponent)
            leax -4,X         ; Reset X to point to 2nd float
            
fadd_opdet: ldaa 0,SP         ; Retrieve the sign information from before.
            adda #1           ; info += 1
            anda #2           ; info &= 2
            tsta              ; if (info != 0) (signs are different)
            bne  fadd_sub     ; Subtract 2nd float from 1st if different signs
fadd_add:   bsr  add_procf    ; Else, add the floats
            bra  fadd_uproc   ; Jump to start unprocessing the floats
fadd_sub:   bsr  sub_procf    ; Branch to subtract processed floats

fadd_uproc: jsr  funproc      ; Unprocess 2nd float (back to IEEE 754)
            ldaa 0,SP         ; Load saved sign information
            cmpa #3           ; If both floats were negative, then...
            beq  fadd_negt    ; Make the 2nd float negative
            cmpa 1,SP         ; Else, if the larger float was negative, then...
            beq  fadd_negt    ; Negate the 2nd float for this case as well
            bra  fadd_end     ; Clean up and restore the 1st float

fadd_negt:  jsr  fsetneg      ; Set 2nd float to be negative
fadd_end:   leax 4,X          ; Set X to point to the 1st float
            jsr  funproc      ; Jump to unprocess the 1st float
            leax -4,X         ; Set X back to the 2nd float
            ldaa 0,SP         ; Load sign information
            anda #1           ; A &= 1
            tsta              ; if (A == 0) (the 1st float was positive)
            beq  fadd_fin     ; Jump to finalize and exit
fadd_ngf1:  leax 4,X          ; Set X to point to 1st float
            jsr  fsetneg      ; Jump to set 1st float negative 
fadd_fin:   leas 2,SP         ; Advance the stack back up to return address
            rts               ; Return to caller

; Subtracts the 2nd float on the stack from the 1st float, replacing the
; 2nd float with the difference, by negating the 2nd float and adding it.
fsub:
            leax 2,SP
            jsr  fnegate
            jsr  fadd
            rts

; Adds 2 48-bit numbers for multiply-accumulator.
add_48:     
            ldab #5
            leax 5,X
            ldaa 0,X
            adda 6,X
            staa 1,X-
a48_acclp:  ldaa 0,X
            adca 6,X
            staa 1,X-
            dbne B,a48_acclp
a48_done:   rts

; Multiplies 2 24-bit processed floats.
mul_procf:
            leas -12,SP
            
            ldd  2,X
            ldy  6,X
            emul
            sty  2,SP
            std  4,SP
            
            clra
            ldab 1,X
            ldy  6,X
            emul
            sty  6,SP
            std  8,SP 
            movw #0,0,SP
            movw #0,10,SP
            pshx
            leax 2,SP
            bsr  add_48
            pulx
            
            ldy  2,X
            clra
            ldab 5,X
            emul
            sty  6,SP
            std  8,SP
            movw #0,10,SP
            pshx
            leax 2,SP
            bsr  add_48
            pulx
            
            ldaa 1,X
            ldab 5,X
            mul
            std  6,SP
            movw #0,8,SP
            movw #0,10,SP
            pshx
            leax 2,SP
            bsr  add_48
            pulx
            
            clra
            ldab 0,X
            subb #$7f
            addb 4,X
            incb
            stab 0,X
            
            movw 0,SP,1,X
            movb 2,SP,3,X
            jsr  frenorm
            leas 12,SP
            rts

; Processes and multiplies two floats.
fmul:
            leas -2,SP
            leax 4,SP
            jsr fabs_sign
            staa 0,SP
            
fmul_proc:  jsr  fproc
            leax 4,X
            jsr  fproc
            leax -4,X
            
fmul_mul:   bsr  mul_procf
fmul_uproc: jsr  funproc
            ldaa 0,SP
            adda #1
            anda #2
            tsta
            beq  fmul_end
fmul_negt:  jsr  fsetneg
fmul_end:   leax 4,X
            jsr  funproc
            leax -4,X
            ldaa 0,SP
            anda #1
            tsta
            beq  fmul_fin
fmul_ngf1:  leax 4,X
            jsr  fsetneg
fmul_fin:   leas 2,SP
            rts

; Multiplies the float pushed by -2 (fast way to get initial Newton-Raphson
; estimate for division).
fmul_m2: 
            leax 2,SP
            jsr  fproc
            ldaa 0,X
            inca
            staa 0,X
            jsr  funproc
            jsr  fnegate
            rts          

; 0: X0, 4: D
fdiv_x1:
			leax 2,SP
			leas -8,SP
			movw 0,X,0,SP
			movw 2,X,2,SP
			movw 4,X,4,SP
			movw 6,X,6,SP
			jsr  fmul			; D*X0
			jsr  fnegate		; -D*X0
			movw #$4000,4,SP
			movw #$0000,6,SP
			jsr  fadd			; 2 - D*X0
         leax 10,SP
			movw 0,X,4,SP
			movw 2,X,6,SP
			jsr  fmul			; X0 * (2 - D*X0)
         leax 10,SP
			movw 0,SP,0,X
			movw 2,SP,2,X
			leas 8,SP
			rts
			
; 0: D	
fdiv_x0:
			leax 2,SP
			leas -4,SP
			movw 0,X,0,SP
			movw 2,X,2,SP
			bsr  fmul_m2		; -2*D
			leas -4,SP
			movw #$403a,0,SP
			movw #$8241,2,SP
			jsr  fadd			; 2.9142 - 2*D
         leax 10,SP
         movw 0,SP,0,X
         movw 2,SP,2,X
			leas 8,SP
			rts

; Scales the dividend and divisor, which are pushed to the stack as arguments.
; 0: D, 4: Q
fdiv_scale:
            leax 2,SP            ; Set X to point to first argument
            leas -1,SP           ; Set SP to allow for 1 temp variable
            jsr  fproc
            ldaa  0,X
            suba #126
            staa 0,SP
            ldaa #126
            staa 0,X
            jsr  funproc
            
            leax 4,X
            jsr fproc
            ldaa 0,X
            suba 0,SP
            staa 0,X
            jsr funproc
            
            leas 1,SP
            rts
            
            
            ;ldd  0,X             ; Get divisor[31:16]
            ;lsld                 ; Shift exp into A
            ;rolb                 ; Shift sign bit into B[0]
            ;suba #126            ; exp = exp - 126
            ;staa 0,SP            ; store in temp space
            ;ldaa #126            ; exp = 126
            ;lsrb                 ; shift sign bit out of B[0]
            ;rora                 ; shift sign bit into A[7]
            ;std  0,X             ; Store new exp  
            ;ldd  4,X             ; Get divisor[15:0]
            ;lsld                 ; Shift exp into A
            ;rolb                 ; Shift sign bit into B[0]
            ;suba 0,SP            ; exp_dividend = exp_dividend - (exp_divisor - 126)
            ;lsrb                 ; Shift sign bit out of B[0]
            ;rora                 ; Shift sign bit into A[7]
            ;std  4,X             ; Store new exp.
            ;leas 1,SP            ; Reset SP to proper place
            ;rts                  ; return


; 0: D, 4: Q			
fdiv:		   leax 2,SP
            leas -9,SP
            movw 0,X,0,SP		   ; 0: D
            movw 2,X,2,SP
            movw 4,X,4,SP		   ; 4: Q
            movw 6,X,6,SP
            movb #4,8,SP		   ; 4 iterations are needed for 32 bits
            bsr  fdiv_scale		; copy D and Q and scale them
			
            leax 0,SP
            leas -4,SP
            movw 0,X,0,SP		   ; 0: D
            movw 2,X,2,SP
            bsr  fdiv_x0		   ; D: X0 = 2.9142 - 2*D
			
do_iters:   jsr  fdiv_x1		   ; X1 = X0 * (2 - D*X0)
            dec  8,SP
            tst  8,SP
            bne  do_iters

do_mult:	   leax 4,SP
            movw 4,X,4,SP
            movw 6,X,6,SP
            jsr  fmul			   ; should return quotient
            leax 15,SP
            movw 0,SP,0,X
            movw 2,SP,2,X
            leas 13,SP
            rts
            
          
            
            

            org RESET_VEC
            dw  EPROM_START
            
